/* 
 * Copyright (C) 2010 Andrew Beekhof <andrew@beekhof.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>

#include "matahari/logging.h"
#include "matahari/mainloop.h"
#include "matahari/services.h"
#include "services_private.h"
#include "sigar.h"

MH_TRACE_INIT_DATA(mh_services);


GHashTable *recurring_actions = NULL;

svc_action_t *services_action_create(
    const char *name, const char *action, int interval, int timeout)
{
    svc_action_t *op = malloc(sizeof(svc_action_t));
    memset(op, 0, sizeof(svc_action_t));

    op->opaque = malloc(sizeof(svc_action_private_t));
    memset(op->opaque, 0, sizeof(svc_action_private_t));
    
    op->rsc = strdup(name);
    op->action = strdup(action);
    op->interval = interval;
    op->timeout = timeout;

    op->rclass = strdup("lsb");
    op->agent = strdup(name);

    op->id = malloc(500);
    snprintf(op->id, 500, "%s_%s_%d", name, action, interval);

    services_os_set_exec(op);
    
    return op;
}

svc_action_t *resources_action_create(
    const char *name, const char *provider, const char *agent,
    const char *action, int interval, int timeout, GHashTable *params)
{
    svc_action_t *op = malloc(sizeof(svc_action_t));
    memset(op, 0, sizeof(svc_action_t));

    op->opaque = malloc(sizeof(svc_action_private_t));
    memset(op->opaque, 0, sizeof(svc_action_private_t));    
    
    op->rsc = strdup(name);
    op->action = strdup(action);
    op->interval = interval;
    op->timeout = timeout;

    op->rclass = strdup("ocf");
    op->agent = strdup(agent);
    op->provider = strdup(provider);

    op->params = params;

    op->id = malloc(500);
    snprintf(op->id, 500, "%s_%s_%d", name, action, interval);

    op->opaque->exec = malloc(500);
    snprintf(op->opaque->exec, 500, "%s/resource.d/%s/%s", OCF_ROOT, provider, agent);

    op->opaque->args[0] = strdup(op->opaque->exec);
    op->opaque->args[1] = strdup(action);
    op->opaque->args[2] = NULL;
    
    return op;
}

void services_action_free(svc_action_t *op)
{
    if(op == NULL) {
	return;
    }
    
    free(op->id);
    free(op->opaque->exec);

    free(op->opaque->args[0]);
    free(op->opaque->args[1]);
    free(op->opaque->args[2]);
    free(op->opaque->args[3]);

    free(op->rsc);
    free(op->action);

    free(op->rclass);
    free(op->agent);
    free(op->provider);

    free(op->stdout_data);
    free(op->stderr_data);

    if(op->params) {
	g_hash_table_destroy(op->params);
	op->params = NULL;
    }

    free(op);
}

gboolean services_action_cancel(const char *name, const char *action, int interval)
{
    svc_action_t* op = NULL;
    gboolean found = FALSE;
    char *id = malloc(500);

    snprintf(id, 500, "%s_%s_%d", name, action, interval);
    op = g_hash_table_lookup(recurring_actions, id);
    free(id);

    if(op) {
	found = TRUE;
	op->opaque->cancel = TRUE;
	mh_debug("Removing %s", op->id);
	if(op->opaque->repeat_timer) {
	    g_source_remove(op->opaque->repeat_timer);
	}
	services_action_free(op);
    }
    
    return found;
}

gboolean
services_action_async(svc_action_t* op, void (*action_callback)(svc_action_t*))
{
    if(action_callback) {
	op->opaque->callback = action_callback;
    }

    if(recurring_actions == NULL) {
	recurring_actions = g_hash_table_new_full(
	    g_str_hash, g_str_equal, NULL, NULL);
    }

    if(op->interval > 0) {
	g_hash_table_replace(recurring_actions, op->id, op);
    }
    
    return services_os_action_execute(op, FALSE);
}

gboolean
services_action_sync(svc_action_t* op)
{
    gboolean rc = services_os_action_execute(op, TRUE);
    mh_trace(" > %s_%s_%d: %s = %d", op->rsc, op->action, op->interval, op->opaque->exec, op->rc);
    if(op->stdout_data) {
	mh_trace(" >  stdout: %s", op->stdout_data);
    }
    if(op->stderr_data) {
	mh_trace(" >  stderr: %s", op->stderr_data);
    }
    return rc;
}

GList *
get_directory_list(const char *root, gboolean files)
{
    return services_os_get_directory_list(root, files);
}

GList *services_list(void) 
{
    return services_os_list();
}

GList *resources_list_ocf_providers(void) 
{
    return resources_os_list_ocf_providers();
}

GList *resources_list_ocf_agents(const char *provider) 
{
    return resources_os_list_ocf_agents(provider);
}

